from functools import wraps

import numpy as np


__all__ = [
    "sphere",
    "rastrigin",
    "bukin",
    "rosenbrock",
    "d1",
]


def sphere(x):
    return np.einsum("ij,ij->i", x, x)


def rastrigin(x, a=10):
    return a * x.shape[1] + (np.square(x) - a * np.cos(2 * np.pi * x)).sum(axis=1)


def bukin(x):
    return 100 * np.sqrt(np.abs(x[:, 1] - 0.01 * np.square(x[:, 0]))) + 0.01 * np.abs(
        x[:, 0] + 10
    )


def rosenbrock(x):
    s = np.zeros(x.shape[0])
    for i in range(x.shape[1] - 1):
        s += 100 * np.square(x[:, i + 1] - np.square(x[:, i])) + np.square(1 - x[:, i])
    return s


def d1(func):
    @wraps(func)
    def _func(obj, list_params, n_iter, **kwargs):
        pa = np.array([x.vals for x in list_params])
        res = func(obj, pa, **kwargs)[:, np.newaxis]
        res = np.repeat(res, len(obj.features), axis=1)
        fe = [obj.features.assign(x).set_error_L2() for x in res]
        return np.array([x.errors for x in fe]), res

    return _func


if __name__ == "__main__":
    X = np.array([[1, 1]])
    aa = rosenbrock(X)
    print(aa)
