import logging
import os
import sys

import treefiles as tf


def load_legends(fname: str):  # -> Tuple[TParams, TFeatures]
    from gcma.io.feature import Features

    d = tf.load_json(fname)
    params = tf.Params(d["params"])
    features = Features(d["features"])
    default_params = tf.Params(d.get("default_params", {}))
    return params, features, default_params


# def r(x, k=2):
#     return None if x is None else round(x, k)


class NoStdout:
    def __init__(self, enable: bool):
        self.enable = enable
        if enable:
            self.old_target = sys.stdout
            sys.stdout = open(os.devnull, "w")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.enable:
            sys.stdout.close()
            sys.stdout = self.old_target


colors = {
    "red": "#d62728",
    "blue": "#1f77b4",
    "green": "#2ca02c",
    "grey": "grey",
    "b": "tab:blue",
    "o": "tab:orange",
    "g": "tab:green",
    "r": "tab:red",
    "p": "tab:purple",
}


log = logging.getLogger(__name__)
